import logging

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from retrying import retry
from typing import List  # noqa: F401

from databuilder import Scoped
from databuilder.filesystem.metadata import FileMetadata

LOGGER = logging.getLogger(__name__)
CLIENT_ERRORS = {'ClientError', 'FileNotFoundError', 'ParamValidationError'}


def is_client_side_error(e):
    # type: (Exception) -> bool
    """
    An method that determines if the error is client side error within FileSystem context
    :param e:
    :return:
    """
    return e.__class__.__name__ in CLIENT_ERRORS


def is_retriable_error(e):
    # type: (Exception) -> bool
    """
    An method that determines if the error is retriable error within FileSystem context
    :param e:
    :return:
    """

    return not is_client_side_error(e)


class FileSystem(Scoped):
    """
    An high level file system, that utilizes Dask File system.
    http://docs.dask.org/en/latest/remote-data-services.html

    All remote call leverages retry against any failure. https://pypi.org/project/retrying/
    """

    # METADATA KEYS
    LAST_UPDATED = 'last_updated'
    SIZE = 'size'

    # CONFIG KEYS
    DASK_FILE_SYSTEM = 'dask_file_system'

    # File metadata that is provided via info(path) method on Dask file system provides a dictionary. As dictionary
    # does not guarantee same key across different implementation, user can provide key mapping.
    FILE_METADATA_MAPPING_KEY = 'file_metadata_mapping'

    default_metadata_mapping = {LAST_UPDATED: 'LastModified',
                                SIZE: 'Size'}
    DEFAULT_CONFIG = ConfigFactory.from_dict({FILE_METADATA_MAPPING_KEY: default_metadata_mapping})

    def init(self,
             conf  # type: ConfigTree
             ):
        # type: (...) -> None
        """
        Initialize Filesystem with DASK file system instance
        Dask file system supports multiple remote storage such as S3, HDFS, Google cloud storage,
        Azure Datalake, etc

        http://docs.dask.org/en/latest/remote-data-services.html
        https://github.com/dask/s3fs
        https://github.com/dask/hdfs3
        ...

        :param conf: hocon config
        :return:
        """
        self._conf = conf.with_fallback(FileSystem.DEFAULT_CONFIG)
        self._dask_fs = self._conf.get(FileSystem.DASK_FILE_SYSTEM)
        self._metadata_key_mapping = self._conf.get(FileSystem.FILE_METADATA_MAPPING_KEY).as_plain_ordered_dict()

    @retry(retry_on_exception=is_retriable_error, stop_max_attempt_number=3, wait_exponential_multiplier=1000,
           wait_exponential_max=5000)
    def ls(self, path):
        # type: (str) -> List[str]
        """
        A scope for the config. Typesafe config supports nested config.
        Scope, string, is used to basically peel off nested config
        :return:
        """
        return self._dask_fs.ls(path)

    @retry(retry_on_exception=is_retriable_error, stop_max_attempt_number=3, wait_exponential_multiplier=1000,
           wait_exponential_max=5000)
    def is_file(self, path):
        # type: (str) -> bool
        contents = self._dask_fs.ls(path)
        return len(contents) == 1 and contents[0] == path

    @retry(retry_on_exception=is_retriable_error, stop_max_attempt_number=3, wait_exponential_multiplier=1000,
           wait_exponential_max=5000)
    def info(self, path):
        # type: (str) -> FileMetadata
        """
        Metadata information about the file. It utilizes _metadata_key_mapping when fetching metadata so that it can
        deal with different keys
        :return:
        """
        metadata_dict = self._dask_fs.info(path)
        fm = FileMetadata(path=path,
                          last_updated=metadata_dict[self._metadata_key_mapping[FileSystem.LAST_UPDATED]],
                          size=metadata_dict[self._metadata_key_mapping[FileSystem.SIZE]])
        return fm

    def get_scope(self):
        # type: () -> str
        return 'filesystem'



import logging
from collections import namedtuple

import google.oauth2.service_account
import google_auth_httplib2
from googleapiclient.discovery import build
import httplib2
from pyhocon import ConfigTree  # noqa: F401
from typing import List, Any  # noqa: F401

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata


DatasetRef = namedtuple('DatasetRef', ['datasetId', 'projectId'])
TableKey = namedtuple('TableKey', ['schema_name', 'table_name'])

LOGGER = logging.getLogger(__name__)


class BigQueryMetadataExtractor(Extractor):

    """ A metadata extractor for bigquery tables, taking the schema metadata
    from the google cloud bigquery API's. This extractor goes through all visible
    datasets in the project identified by project_id and iterates over all tables
    it finds. A separate account is configurable through the key_path parameter,
    which should point to a valid json file corresponding to a service account.

    This extractor supports nested columns, which are delimited by a dot (.) in the
    column name.
    """

    PROJECT_ID_KEY = 'project_id'
    KEY_PATH_KEY = 'key_path'
    PAGE_SIZE_KEY = 'page_size'
    FILTER_KEY = 'filter'
    _DEFAULT_SCOPES = ['https://www.googleapis.com/auth/bigquery.readonly', ]
    DEFAULT_PAGE_SIZE = 300
    NUM_RETRIES = 3
    DATE_LENGTH = 8

    def init(self, conf):
        # type: (ConfigTree) -> None
        self.key_path = conf.get_string(BigQueryMetadataExtractor.KEY_PATH_KEY, None)
        self.project_id = conf.get_string(BigQueryMetadataExtractor.PROJECT_ID_KEY)
        self.pagesize = conf.get_int(
            BigQueryMetadataExtractor.PAGE_SIZE_KEY,
            BigQueryMetadataExtractor.DEFAULT_PAGE_SIZE)
        self.filter = conf.get_string(BigQueryMetadataExtractor.FILTER_KEY, '')

        if self.key_path:
            credentials = (
                google.oauth2.service_account.Credentials.from_service_account_file(
                    self.key_path, scopes=BigQueryMetadataExtractor._DEFAULT_SCOPES))
        else:
            credentials, _ = google.auth.default(scopes=BigQueryMetadataExtractor._DEFAULT_SCOPES)

        http = httplib2.Http()
        authed_http = google_auth_httplib2.AuthorizedHttp(credentials, http=http)
        self.bigquery_service = build('bigquery', 'v2', http=authed_http, cache_discovery=False)
        self.datasets = self._retrieve_datasets()
        self.iter = iter(self._iterate_over_tables())
        self.grouped_tables = set([])

    def extract(self):
        # type: () -> Any
        try:
            return next(self.iter)
        except StopIteration:
            return None

    def _iterate_over_tables(self):
        # type: () -> Any
        for dataset in self.datasets:
            for entry in self._retrieve_tables(dataset):
                yield(entry)

    def _retrieve_datasets(self):
        # type: () -> List[DatasetRef]
        datasets = []
        for page in self._page_dataset_list_results():
            if 'datasets' not in page:
                continue

            for dataset in page['datasets']:
                dataset_ref = dataset['datasetReference']
                ref = DatasetRef(**dataset_ref)
                datasets.append(ref)

        return datasets

    def _page_dataset_list_results(self):
        # type: () -> Any
        response = self.bigquery_service.datasets().list(
            projectId=self.project_id,
            all=False,  # Do not return hidden datasets
            filter=self.filter,
            maxResults=self.pagesize).execute(
                num_retries=BigQueryMetadataExtractor.NUM_RETRIES)

        while response:
            yield response

            if 'nextPageToken' in response:
                response = self.bigquery_service.datasets().list(
                    projectId=self.project_id,
                    all=True,
                    filter=self.filter,
                    pageToken=response['nextPageToken']).execute(
                        num_retries=BigQueryMetadataExtractor.NUM_RETRIES)
            else:
                response = None

    def _retrieve_tables(self, dataset):
        # type: () -> Any
        for page in self._page_table_list_results(dataset):
            if 'tables' not in page:
                continue

            for table in page['tables']:
                tableRef = table['tableReference']

                table_id = tableRef['tableId']

                # BigQuery tables that have 8 digits as last characters are
                # considered date range tables and are grouped together in the UI.
                # ( e.g. ga_sessions_20190101, ga_sessions_20190102, etc. )
                last_eight_chars = table_id[-BigQueryMetadataExtractor.DATE_LENGTH:]
                if last_eight_chars.isdigit():
                    # If the last eight characters are digits, we assume the table is of a table date range type
                    # and then we only need one schema definition
                    table_prefix = table_id[:-BigQueryMetadataExtractor.DATE_LENGTH]
                    if table_prefix in self.grouped_tables:
                        # If one table in the date range is processed, then ignore other ones
                        # (it adds too much metadata)
                        continue

                    table_id = table_prefix
                    self.grouped_tables.add(table_prefix)

                table = self.bigquery_service.tables().get(
                    projectId=tableRef['projectId'],
                    datasetId=tableRef['datasetId'],
                    tableId=tableRef['tableId']).execute(num_retries=BigQueryMetadataExtractor.NUM_RETRIES)

                # BigQuery tables also have interesting metadata about partitioning
                # data location (EU/US), mod/create time, etc... Extract that some other time?
                schema = table['schema']
                cols = []
                if 'fields' in schema:
                    total_cols = 0
                    for column in schema['fields']:
                        total_cols = self._iterate_over_cols('', column, cols, total_cols + 1)

                table_meta = TableMetadata(
                    database='bigquery',
                    cluster=tableRef['projectId'],
                    schema_name=tableRef['datasetId'],
                    name=table_id,
                    description=table.get('description', ''),
                    columns=cols,
                    is_view=table['type'] == 'VIEW')

                yield(table_meta)

    def _iterate_over_cols(self, parent, column, cols, total_cols):
        # type: (str, str, List[ColumnMetadata()], int) -> int
        if len(parent) > 0:
            col_name = '{parent}.{field}'.format(parent=parent, field=column['name'])
        else:
            col_name = column['name']

        if column['type'] == 'RECORD':
            col = ColumnMetadata(
                name=col_name,
                description=column.get('description', ''),
                col_type=column['type'],
                sort_order=total_cols)
            cols.append(col)
            total_cols += 1
            for field in column['fields']:
                total_cols = self._iterate_over_cols(col_name, field, cols, total_cols)
            return total_cols
        else:
            col = ColumnMetadata(
                name=col_name,
                description=column.get('description', ''),
                col_type=column['type'],
                sort_order=total_cols)
            cols.append(col)
            return total_cols + 1

    def _page_table_list_results(self, dataset):
        # type: (DatasetRef) -> Any
        response = self.bigquery_service.tables().list(
            projectId=dataset.projectId,
            datasetId=dataset.datasetId,
            maxResults=self.pagesize).execute(
                num_retries=BigQueryMetadataExtractor.NUM_RETRIES)

        while response:
            yield response

            if 'nextPageToken' in response:
                response = self.bigquery_service.tables().list(
                    projectId=dataset.projectId,
                    datasetId=dataset.datasetId,
                    maxResults=self.pagesize,
                    pageToken=response['nextPageToken']).execute(
                        num_retries=BigQueryMetadataExtractor.NUM_RETRIES)
            else:
                response = None

    def get_scope(self):
        # type: () -> str
        return 'extractor.bigquery_table_metadata'



import logging
import six
from collections import namedtuple

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from typing import Iterator, Union, Dict, Any  # noqa: F401
from unidecode import unidecode

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata
from itertools import groupby


TableKey = namedtuple('TableKey', ['schema_name', 'table_name'])

LOGGER = logging.getLogger(__name__)


class SnowflakeMetadataExtractor(Extractor):
    """
    Extracts Snowflake table and column metadata from underlying meta store database using SQLAlchemyExtractor.
    Requirements:
        snowflake-connector-python
        snowflake-sqlalchemy
    """
    # SELECT statement from snowflake information_schema to extract table and column metadata
    SQL_STATEMENT = """
    SELECT
        lower(c.column_name) AS col_name,
        c.comment AS col_description,
        lower(c.data_type) AS col_type,
        lower(c.ordinal_position) AS col_sort_order,
        lower(c.table_catalog) AS database,
        lower({cluster_source}) AS cluster,
        lower(c.table_schema) AS schema_name,
        lower(c.table_name) AS name,
        t.comment AS description,
        decode(lower(t.table_type), 'view', 'true', 'false') AS is_view
    FROM
        {database}.INFORMATION_SCHEMA.COLUMNS AS c
    LEFT JOIN
        {database}.INFORMATION_SCHEMA.TABLES t
            ON c.TABLE_NAME = t.TABLE_NAME
            AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
    {where_clause_suffix};
    """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster_key'
    USE_CATALOG_AS_CLUSTER_NAME = 'use_catalog_as_cluster_name'
    DATABASE_KEY = 'database_key'

    # Default values
    DEFAULT_CLUSTER_NAME = 'master'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {WHERE_CLAUSE_SUFFIX_KEY: ' ',
         CLUSTER_KEY: DEFAULT_CLUSTER_NAME,
         USE_CATALOG_AS_CLUSTER_NAME: True,
         DATABASE_KEY: 'prod'}
    )

    def init(self, conf):
        # type: (ConfigTree) -> None
        conf = conf.with_fallback(SnowflakeMetadataExtractor.DEFAULT_CONFIG)
        self._cluster = '{}'.format(conf.get_string(SnowflakeMetadataExtractor.CLUSTER_KEY))

        if conf.get_bool(SnowflakeMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME):
            cluster_source = "c.table_catalog"
        else:
            cluster_source = "'{}'".format(self._cluster)

        self._database = conf.get_string(SnowflakeMetadataExtractor.DATABASE_KEY)
        if six.PY2:
            self._database = self._database.encode('utf-8', 'ignore')

        self.sql_stmt = SnowflakeMetadataExtractor.SQL_STATEMENT.format(
            where_clause_suffix=conf.get_string(SnowflakeMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY),
            cluster_source=cluster_source,
            database=self._database
        )

        LOGGER.info('SQL for snowflake metadata: {}'.format(self.sql_stmt))

        self._alchemy_extractor = SQLAlchemyExtractor()
        sql_alch_conf = Scoped.get_scoped_conf(conf, self._alchemy_extractor.get_scope())\
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt}))

        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter = None  # type: Union[None, Iterator]

    def extract(self):
        # type: () -> Union[TableMetadata, None]
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.snowflake'

    def _get_extract_iter(self):
        # type: () -> Iterator[TableMetadata]
        """
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        """
        for key, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                columns.append(ColumnMetadata(
                               row['col_name'],
                               unidecode(row['col_description']) if row['col_description'] else None,
                               row['col_type'],
                               row['col_sort_order'])
                               )

            yield TableMetadata(self._database, last_row['cluster'],
                                last_row['schema_name'],
                                last_row['name'],
                                unidecode(last_row['description']) if last_row['description'] else None,
                                columns)

    def _get_raw_extract_iter(self):
        # type: () -> Iterator[Dict[str, Any]]
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()

    def _get_table_key(self, row):
        # type: (Dict[str, Any]) -> Union[TableKey, None]
        """
        Table key consists of schema and table name
        :param row:
        :return:
        """
        if row:
            return TableKey(schema_name=row['schema_name'], table_name=row['name'])

        return None



from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE

from databuilder.models.table_metadata import TableMetadata


class TableSource(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Hive table source model.
    """
    LABEL = 'Source'
    KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/_source'
    SOURCE_TABLE_RELATION_TYPE = 'SOURCE_OF'
    TABLE_SOURCE_RELATION_TYPE = 'SOURCE'

    def __init__(self,
                 db_name,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 cluster,  # type: str
                 source,  # type: str
                 source_type='github',  # type: str
                 ):
        # type: (...) -> None
        self.db = db_name.lower()
        self.schema = schema_name.lower()
        self.table = table_name.lower()

        self.cluster = cluster.lower() if cluster else 'gold'
        # source is the source file location
        self.source = source
        self.source_type = source_type
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_source_model_key(self):
        # type: (...) -> str
        return TableSource.KEY_FORMAT.format(db=self.db,
                                             cluster=self.cluster,
                                             schema=self.schema,
                                             tbl=self.table)

    def get_metadata_model_key(self):
        # type: (...) -> str
        return '{db}://{cluster}.{schema}/{table}'.format(db=self.db,
                                                          cluster=self.cluster,
                                                          schema=self.schema,
                                                          table=self.table)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = [{
            NODE_KEY: self.get_source_model_key(),
            NODE_LABEL: TableSource.LABEL,
            'source': self.source,
            'source_type': self.source_type
        }]
        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation map between owner record with original hive table
        :return:
        """
        results = [{
            RELATION_START_KEY: self.get_source_model_key(),
            RELATION_START_LABEL: TableSource.LABEL,
            RELATION_END_KEY: self.get_metadata_model_key(),
            RELATION_END_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_TYPE: TableSource.SOURCE_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableSource.TABLE_SOURCE_RELATION_TYPE
        }]
        return results

    def __repr__(self):
        # type: () -> str
        return 'TableSource({!r}, {!r}, {!r}, {!r}, {!r})'.format(self.db,
                                                                  self.cluster,
                                                                  self.schema,
                                                                  self.table,
                                                                  self.source)



import json
from abc import ABCMeta


class ElasticsearchDocument:
    """
    Base class for ElasticsearchDocument
    Each different resource ESDoc will be a subclass
    """
    __metaclass__ = ABCMeta

    def to_json(self):
        # type: () -> str
        """
        Convert object to json
        :return:
        """
        obj_dict = {k: v for k, v in sorted(self.__dict__.items())}
        data = json.dumps(obj_dict) + "\n"
        return data



from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)

from databuilder.models.table_metadata import TableMetadata, DESCRIPTION_NODE_LABEL


class TestColumnMetadata(Neo4jCsvSerializable):
    COLUMN_NODE_LABEL = 'Column'
    COLUMN_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}'
    COLUMN_NAME = 'name'
    COLUMN_TYPE = 'type'
    COLUMN_ORDER = 'sort_order'
    COLUMN_DESCRIPTION = 'description'
    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/_description'

    # pair of nodes makes relationship where name of variable represents order of relationship.
    COL_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_COL_RELATION_TYPE = 'DESCRIPTION_OF'

    def __init__(self,
                 name,  # type: str
                 description,  # type: Union[str, None]
                 col_type,  # type: str
                 sort_order,  # type: int
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 table_description,  # type: str
                 ):
        # type: (...) -> None
        """
        TODO: Add stats
        :param name:
        :param description:
        :param col_type:
        :param sort_order:
        """
        self.name = name
        self.description = description
        self.type = col_type
        self.sort_order = sort_order
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.table_name = table_name
        self.table_desc = table_description

        # currently we don't consider nested partitions
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def _get_col_key(self):
        # type: (TestColumnMetadata) -> str
        return TestColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                           cluster=self.cluster,
                                                           schema=self.schema_name,
                                                           tbl=self.table_name,
                                                           col=self.name)

    def _get_col_description_key(self):
        # type: (TestColumnMetadata) -> str
        return TestColumnMetadata.COLUMN_DESCRIPTION_FORMAT.format(db=self.database,
                                                                   cluster=self.cluster,
                                                                   schema=self.schema_name,
                                                                   tbl=self.table_name,
                                                                   col=self.name)

    def _get_table_key(self):
        # type: () -> str
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                     cluster=self.cluster,
                                                     schema=self.schema_name,
                                                     tbl=self.table_name)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = [{
            NODE_LABEL: TestColumnMetadata.COLUMN_NODE_LABEL,
            NODE_KEY: self._get_col_key(),
            TestColumnMetadata.COLUMN_NAME: self.name,
            TestColumnMetadata.COLUMN_TYPE: self.type,
            TestColumnMetadata.COLUMN_ORDER: self.sort_order
        }]

        if self.description:
            results.append({
                NODE_LABEL: DESCRIPTION_NODE_LABEL,
                NODE_KEY: self._get_col_description_key(),
                TestColumnMetadata.COLUMN_DESCRIPTION: self.description
            })

        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation map between watermark record with original hive table
        :return:
        """

        results = [{
            RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_END_LABEL: TestColumnMetadata.COLUMN_NODE_LABEL,
            RELATION_START_KEY: self._get_table_key(),
            RELATION_END_KEY: self._get_col_key(),
            RELATION_TYPE: TableMetadata.TABLE_COL_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableMetadata.COL_TABLE_RELATION_TYPE
        }]

        if self.description:
            results.append({
                RELATION_START_LABEL: self.COLUMN_NODE_LABEL,
                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_col_key(),
                RELATION_END_KEY: self._get_col_description_key(),
                RELATION_TYPE: self.COL_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: self.DESCRIPTION_COL_RELATION_TYPE
            })

        return results

    def __repr__(self):
        # type: () -> str
        return 'ColumnMetadata({!r}, {!r}, {!r}, {!r})'.format(self.name,
                                                               self.description,
                                                               self.type,
                                                               self.sort_order)



class PrestoQueryLogs:
    """
    Presto Query logs model.
    Sql result has one row per presto query.
    """
    def __init__(self,
                 user,  # type: str
                 query_text,  # type: str
                 occurred_at  # type: str
                 ):
        # type: (...) -> None
        self.user = user
        self.query_text = query_text
        self.occurred_at = occurred_at



import logging
from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker
from typing import Iterable, List  # noqa: F401

from databuilder.sql_parser.usage.column import Column, Table, remove_double_quotes
from databuilder.sql_parser.usage.presto.antlr_generated.SqlBaseLexer import SqlBaseLexer
from databuilder.sql_parser.usage.presto.antlr_generated.SqlBaseListener import SqlBaseListener
from databuilder.sql_parser.usage.presto.antlr_generated.SqlBaseParser import SqlBaseParser


LOGGER = logging.getLogger(__name__)


class ColumnUsageListener(SqlBaseListener):
    """
    ColumnUsageListener that inherits Antlr generated SqlBaseListener so that it can extract column and table usage
    while ColumnUsageListener walks the parsing tree.

    (Method name is from Antlr generated SqlBaseListener where it does not follow python convention)

    Basic idea of column extraction is to look at SELECT statement as two parts.
     1. processing columns: Columns in SELECT clause. (SELECT foo, bar )
     2. processed columns: Columns in FROM clause. (FROM foobar or FROM (SELECT foo, bar from foobar) )

    Overall, we'd like to retrieve processing column. Thus, the problem we need to solve is basically based on
    processed column, finalize processing column by get the necessary info (such as table, schema) from processed
    column.
    """
    def __init__(self):
        # type: () -> None
        self.processed_cols = []  # type: List[Column]
        self._processing_cols = []  # type: List[Column]
        self._current_col = None  # type: Column
        self._stack = []  # type: List[Column]

    def exitColumnReference(self,
                            ctx  # type: SqlBaseParser.ColumnReferenceContext
                            ):
        # type: (...) -> None

        """
        Call back method for column that does not have table indicator
        :param ctx:
        :return:
        """
        self._current_col = Column(ctx.getText())

    def exitDereference(self,
                        ctx  # type: SqlBaseParser.DereferenceContext
                        ):
        # type: (...) -> None
        """
        Call back method for column with table indicator e.g: foo.bar
        :param ctx:
        :return:
        """
        self._current_col = Column(ctx.identifier().getText(),
                                   table=Table(ctx.base.getText()))

    def exitSelectSingle(self,
                         ctx  # type: SqlBaseParser.SelectSingleContext
                         ):
        # type: (...) -> None
        """
        Call back method for select single column. This is to distinguish
        between columns for SELECT statement and columns for something else
        such as JOIN statement
        :param ctx:
        :return:
        """
        if not self._current_col:
            return

        if ctx.identifier():
            self._current_col.col_alias = remove_double_quotes(ctx.identifier().getText())

        self._processing_cols.append(self._current_col)
        self._current_col = None

    def exitSelectAll(self,
                      ctx  # type: SqlBaseParser.SelectAllContext
                      ):
        # type: (...) -> None
        """
        Call back method for select ALL column.
        :param ctx:
        :return:
        """
        self._current_col = Column('*')
        if ctx.qualifiedName():
            self._current_col.table = Table(ctx.qualifiedName().getText())

        self._processing_cols.append(self._current_col)
        self._current_col = None

    def exitTableName(self,
                      ctx  # type: SqlBaseParser.TableNameContext
                      ):
        # type: (...) -> None
        """
        Call back method for table name
        :param ctx:
        :return:
        """
        table_name = ctx.getText()
        table = Table(table_name)
        if '.' in table_name:
            db_tbl = table_name.split('.')
            table = Table(db_tbl[len(db_tbl) - 1],
                          schema=db_tbl[len(db_tbl) - 2])

        self._current_col = Column('*', table=table)

    def exitAliasedRelation(self,
                            ctx  # type: SqlBaseParser.AliasedRelationContext
                            ):
        # type: (...) -> None
        """
        Call back method for table alias
        :param ctx:
        :return:
        """
        if not ctx.identifier():
            return

        # Table alias for column
        if self._current_col and self._current_col.table:
            self._current_col.table.alias = remove_double_quotes(ctx.identifier().getText())
            return

        # Table alias for inner SQL
        for col in self.processed_cols:
            col.table.alias = remove_double_quotes(ctx.identifier().getText())

    def exitRelationDefault(self,
                            ctx  # type: SqlBaseParser.RelationDefaultContext
                            ):
        # type: (...) -> None
        """
        Callback method when exiting FROM clause. Here we are moving processing columns to processed
        to processed
        :param ctx:
        :return:
        """
        if not self._current_col:
            return

        self.processed_cols.append(self._current_col)
        self._current_col = None

    def enterQuerySpecification(self,
                                ctx  # type: SqlBaseParser.QuerySpecificationContext
                                ):
        # type: (...) -> None
        """
        Callback method for Query specification. For nested SELECT
        statement, it will store previous processing column to stack.
        :param ctx:
        :return:
        """
        if not self._processing_cols:
            return

        self._stack.append(self._processing_cols)
        self._processing_cols = []

    def exitQuerySpecification(self,
                               ctx  # type: SqlBaseParser.QuerySpecificationContext
                               ):
        # type: (...) -> None
        """
        Call back method for Query specification. It merges processing
        columns with processed column
        :param ctx:
        :return:
        """

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug('processing_cols: {}'.format(self._processing_cols))
            LOGGER.debug('processed_cols: {}'.format(self.processed_cols))

        result = []

        for col in self._processing_cols:
            for resolved in Column.resolve(col, self.processed_cols):
                result.append(resolved)

        self.processed_cols = result
        self._processing_cols = []
        if self._stack:
            self._processing_cols = self._stack.pop()

        self._current_col = None

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug('done processing_cols: {}'.format(self._processing_cols))
            LOGGER.debug('done processed_cols: {}'.format(self.processed_cols))


class ColumnUsageProvider(object):
    def __init__(self):
        # type: () -> None
        pass

    @classmethod
    def get_columns(cls, query):
        # type: (str) -> Iterable[Column]
        """
        Using presto Grammar, instantiate Parsetree, attach ColumnUsageListener to tree and walk the tree.
        Once finished walking the tree, listener will have selected columns and return them.
        :param query:
        :return:
        """

        query = query.rstrip(';').upper() + "\n"
        lexer = SqlBaseLexer(InputStream(query))
        parser = SqlBaseParser(CommonTokenStream(lexer))
        parse_tree = parser.singleStatement()

        listener = ColumnUsageListener()
        walker = ParseTreeWalker()
        walker.walk(listener, parse_tree)

        return listener.processed_cols



import abc

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from typing import Any  # noqa: F401


class Loader(Scoped):
    """
    A loader loads to the destination or to the staging area
    """
    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    @abc.abstractmethod
    def load(self, record):
        # type: (Any) -> None
        pass

    def get_scope(self):
        # type: () -> str
        return 'loader'




import copy
import logging

import six
from typing import Union, Iterable, Optional  # noqa: F401

LOGGER = logging.getLogger(__name__)


def remove_double_quotes(val):
    # type: (Union[str, None]) -> Union[str, None]
    """
    Often times, column name, table name or any identifier comes with double quoted. This method will remove double
    quotes.
    :param val:
    :return:
    """
    if not val:
        return val

    if six.PY2 and isinstance(val, six.text_type):
        val = val.encode('utf-8', 'ignore')

    if val.startswith('"') and val.endswith('"'):
        return val[1:-1]
    return val


class Column(object):
    """
    Column for usage.
    """
    def __init__(self, name, table=None, col_alias=None):
        # type: (str, Union[Table, None], Union[str, None]) -> None
        self.col_name = remove_double_quotes(name)
        self.table = table
        self.col_alias = remove_double_quotes(col_alias)

    def __repr__(self):
        # type: () -> str
        return 'Column(name={!r}, table={!r}, col_alias={!r})'.format(self.col_name, self.table, self.col_alias)

    def resolve_col_name(self, col_name):
        # type: (Union[str, None]) -> Union[str, None]
        """
        Resolve column name for currently processing column.
        e.g: SELECT bar from (SELECT foo as bar FROM foobar)
        Above case, bar is trying to resolve with column foo that has alias bar. In this case, it will resolve to foo,
        as that is the actual column name.
        :param col_name:
        :return:
        """
        if self.col_name == '*':
            return col_name

        if col_name == self.col_alias or col_name == self.col_name:
            return self.col_name

        return None

    @staticmethod
    def resolve(select_col, from_cols):
        # type: (Column, Iterable[Column]) -> Iterable[Column]
        """
        Resolve processing column with processed columns. Processing column is the one from SELECT clause where it
        does not have table information where it can optionally have table alias in front of column (foo.bar)
        Processed columns are already resolved columns that it has column with table with it.

        Resolving processing columns with processed columns means having processing columns choose correct processed
        columns. The result is proper column name with table name.

        e.g1: SELECT foo from foobar
         - processing column: foo
         - processed column: all columns from foobar
         --> result: column 'foo' from 'foobar' table

        e.g2: SELECT foo from (SELECT foo, bar FROM foobar)
         - processing column: foo
         - processed column: foo and bar columns from foobar
         --> result: column 'foo' from 'foobar' table

        :param select_col: column from select clause
        :param from_cols: column from 'from' clause
        :return: List of columns
        """
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug('select_col: {}'.format(select_col))
            LOGGER.debug('from_cols: {}'.format(from_cols))

        if select_col.col_name != '*':
            return Column.resolve_named_column(select_col, from_cols)

        return Column._resolve_all_column(select_col, from_cols)

    @staticmethod
    def resolve_named_column(select_col, from_cols):
        # type: (Column, Iterable[Column]) -> Iterable[Column]
        """
        SELECT clause where column has name (not *)
        e.g: SELECT foo, bar FROM foobar.

        :param select_col: column from select clause
        :param from_cols: column from 'from' clause
        :return:
        """
        # column name is defined and has table alias. (SELECT foo.bar)
        if select_col.table:
            for processed_col in from_cols:
                # resolve column name as processing column name can be alias
                col_name = processed_col.resolve_col_name(select_col.col_name)
                if col_name and processed_col.table:
                    # resolve table name using alias
                    table = processed_col.table.resolve_table(select_col.table.name)
                    if table:
                        alias = Column.get_column_alias(select_col=select_col, from_col=processed_col)
                        col = Column(col_name, table=table, col_alias=alias)
                        return [col]
            raise Exception('Cannot found case 1. select_col: {} , from_cols: {}'
                            .format(select_col, from_cols))
        # col name defined but no table. (SELECT foo)
        else:
            sub_result = []
            # Using column name only to find a match from processed column.
            # Note that we can have a column name with multiple table as a result. This is the case that SQL engine
            # resolves ambiguity by looking into actual columns in table. Here we use OrTable so that later on it
            #  can be disambiguated.
            for processed_col in from_cols:
                col_name = processed_col.resolve_col_name(select_col.col_name)
                if col_name:
                    col = copy.deepcopy(processed_col)
                    col.col_name = col_name
                    alias = Column.get_column_alias(select_col=select_col, from_col=processed_col)
                    col.col_alias = alias
                    sub_result.append(col)

            if not sub_result:
                raise Exception('Cannot find case 2. select_col: {} , from_cols: {}'
                                .format(select_col, from_cols))
            if len(sub_result) == 1:
                return sub_result

            tables = []
            for col in sub_result:
                tables.append(copy.deepcopy(col.table))
            col = sub_result[0]
            col.table = OrTable(tables)
            return [col]

    @staticmethod
    def get_column_alias(select_col, from_col):
        # type: (Column, Column) -> str
        """
        Use processing column alias if not null.
        :param select_col: column from select clause
        :param from_col: column from 'from' clause
        :return:
        """
        return select_col.col_alias if select_col.col_alias else from_col.col_alias

    @staticmethod
    def _resolve_all_column(processing_col, processed_cols):
        # type: (Column, Iterable[Column]) -> Iterable[Column]
        """
        SELECT statement where column is '*'
        e.g: SELECT * FROM foobar;
        :param processed_cols:
        :return:
        """
        if processing_col.table:
            result = []
            # Select whatever we have in processed where it just need to match table
            for processed_col in processed_cols:
                if processed_col.table:
                    table = processed_col.table.resolve_table(processing_col.table.name)
                if table:
                    col = copy.deepcopy(processed_col)
                    col.table = table
                    result.append(col)
            if not result:
                raise Exception('Cannot find case 3. select_col: {} , from_cols: {}'
                                .format(processing_col, processed_cols))
            return result

        # SELECT * case
        else:
            result = []
            for processed_col in processed_cols:
                result.append(copy.deepcopy(processed_col))
            if not result:
                raise Exception('Cannot find case 4. select_col: {} , from_cols: {}'
                                .format(processing_col, processed_cols))
            return result


class Table(object):
    """
    Table class for usage
    """
    def __init__(self, name, schema=None, alias=None):
        # type: (str, Union[str, None], Union[str, None]) -> None
        self.name = remove_double_quotes(name)
        self.schema = remove_double_quotes(schema)
        self.alias = remove_double_quotes(alias)

    def resolve_table(self, select_table_name):
        # type: (Union[str, None]) -> Union[Table, None]
        """
        If processing_table_name matches table return table instance
        :param select_table_name: table in select clause
        :return:
        """
        if select_table_name == self.alias or select_table_name == self.name:
            return self
        return None

    def __repr__(self):
        # type: () -> str
        return 'Table(name={!r}, schema={!r}, alias={!r})'.format(self.name, self.schema, self.alias)


class OrTable(Table):
    """
    Table that holds multiple table. This is for ambiguous case.
    For example, "SELECT a, b FROM foo JOIN bar USING c" statement does not tell if column a is from foo or bar.
    Thus, column a is either from table foo or bar and this class represent this problem.
    """
    def __init__(self, tables):
        # type: (Iterable[Optional[Table]]) -> None
        self.tables = tables

    def resolve_table(self, select_table_name):
        # type: (str) -> Optional[Table]
        """
        If any of term matches with table return it
        :param select_table_name:
        :return:
        """
        for table in self.tables:
            if isinstance(table, OrTable):
                result = table.resolve_table(select_table_name)
                if result:
                    return result
                continue

            if select_table_name == table.alias or select_table_name == table.name:
                return table
        return None

    def __repr__(self):
        # type: () -> str
        return 'OrTable(tables={!r})'.format(self.tables)



from pyhocon import ConfigTree  # noqa: F401
from typing import Dict, Optional  # noqa: F401

from databuilder.transformer.base_transformer import Transformer
from databuilder.models.table_column_usage import ColumnReader, TableColumnUsage
from databuilder.extractor.bigquery_usage_extractor import TableColumnUsageTuple


class BigqueryUsageTransformer(Transformer):

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Transformer to convert TableColumnUsageTuple data to bigquery usage data
        which can be uploaded to Neo4j
        """
        self.conf = conf

    def transform(self, record):
        # type: (Dict) -> Optional[TableColumnUsage]
        if not record:
            return None

        (key, count) = record

        if not isinstance(key, TableColumnUsageTuple):
            raise Exception("BigqueryUsageTransformer expects record of type TableColumnUsageTuple")

        col_readers = []
        col_readers.append(ColumnReader(database=key.database,
                                        cluster=key.cluster,
                                        schema=key.schema,
                                        table=key.table,
                                        column=key.column,
                                        user_email=key.email,
                                        read_count=count))

        return TableColumnUsage(col_readers=col_readers)

    def get_scope(self):
        # type: () -> str
        return 'transformer.bigquery_usage'



try:
    import copy_reg
except Exception:
    import copyreg as copy_reg

import logging
import types
from multiprocessing.pool import Pool, TimeoutError

from pyhocon import ConfigTree  # noqa: F401
from typing import Any, Optional, List, Iterable  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.hive_table_metadata_extractor import HiveTableMetadataExtractor
from databuilder.models.table_column_usage import TableColumnUsage, ColumnReader
from databuilder.sql_parser.usage.column import OrTable, Table  # noqa: F401
from databuilder.sql_parser.usage.presto.column_usage_provider import ColumnUsageProvider
from databuilder.transformer.base_transformer import Transformer

LOGGER = logging.getLogger(__name__)


def _pickle_method(m):
    """
    Pool needs to pickle method in order to pass it to separate process. This method is to define how to pickle method.
    """
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.func_name)
    else:
        return getattr, (m.im_self, m.im_func.func_name)


copy_reg.pickle(types.MethodType, _pickle_method)


class SqlToTblColUsageTransformer(Transformer):
    """
    Note that it currently only supports Presto SQL.
    A SQL to usage transformer where it transforms to ColumnReader that has column, user, count.
    Currently it's collects on table level that column on same table will be de-duped.
    In many cases, "from" clause does not contain schema and this will be fetched via table name -> schema name mapping
    which it gets from Hive metastore. (Naming collision is disregarded as it needs column level to disambiguate)

    Currently, ColumnUsageProvider could hang on certain SQL statement and as a short term solution it will timeout
    processing statement at 10 seconds.
    """
    # Config key
    DATABASE_NAME = 'database'
    CLUSTER_NAME = 'cluster'
    SQL_STATEMENT_ATTRIBUTE_NAME = 'sql_stmt_attribute_name'
    USER_EMAIL_ATTRIBUTE_NAME = 'user_email_attribute_name'
    COLUMN_EXTRACTION_TIMEOUT_SEC = 'column_extraction_timeout_seconds'
    LOG_ALL_EXTRACTION_FAILURES = 'log_all_extraction_failures'

    total_counts = 0
    failure_counts = 0

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._conf = conf
        self._database = conf.get_string(SqlToTblColUsageTransformer.DATABASE_NAME)
        self._cluster = conf.get_string(SqlToTblColUsageTransformer.CLUSTER_NAME, 'gold')
        self._sql_stmt_attr = conf.get_string(SqlToTblColUsageTransformer.SQL_STATEMENT_ATTRIBUTE_NAME)
        self._user_email_attr = conf.get_string(SqlToTblColUsageTransformer.USER_EMAIL_ATTRIBUTE_NAME)
        self._tbl_to_schema_mapping = self._create_schema_by_table_mapping()
        self._worker_pool = Pool(processes=1)
        self._time_out_sec = conf.get_int(SqlToTblColUsageTransformer.COLUMN_EXTRACTION_TIMEOUT_SEC, 10)
        LOGGER.info('Column extraction timeout: {} seconds'.format(self._time_out_sec))
        self._log_all_extraction_failures = conf.get_bool(SqlToTblColUsageTransformer.LOG_ALL_EXTRACTION_FAILURES,
                                                          False)

    def transform(self, record):
        # type: (Any) -> Optional[TableColumnUsage]
        SqlToTblColUsageTransformer.total_counts += 1

        stmt = getattr(record, self._sql_stmt_attr)
        email = getattr(record, self._user_email_attr)

        result = []  # type: List[ColumnReader]
        try:
            columns = self._worker_pool.apply_async(ColumnUsageProvider.get_columns, (stmt,)).get(self._time_out_sec)
            # LOGGER.info('Statement: {} ---> columns: {}'.format(stmt, columns))
        except TimeoutError:
            SqlToTblColUsageTransformer.failure_counts += 1
            LOGGER.exception('Timed out while getting column usage from query: {}'.format(stmt))
            LOGGER.info('Killing the thread.')
            self._worker_pool.terminate()
            self._worker_pool = Pool(processes=1)
            LOGGER.info('Killed the thread.')
            return None
        except Exception:
            SqlToTblColUsageTransformer.failure_counts += 1
            if self._log_all_extraction_failures:
                LOGGER.exception('Failed to get column usage from query: {}'.format(stmt))
            return None

        # Dedupe is needed to make it table level. TODO: Remove this once we are at column level
        dedupe_tuples = set()  # type: set
        for col in columns:
            sub_result = self._get_col_readers(table=col.table,
                                               stmt=stmt,
                                               email=email,
                                               dedupe_tuples=dedupe_tuples)
            result.extend(sub_result)

        if not result:
            return None

        return TableColumnUsage(col_readers=result)

    def _get_col_readers(self,
                         table,  # type: Table
                         stmt,  # type: str
                         email,  # type: str
                         dedupe_tuples  # type: set
                         ):
        # type: (...) -> Iterable[ColumnReader]
        """
        Using table information, produce column reader with de-duping same table as it's from same statement
        :param table:
        :param stmt:
        :param email:
        :param dedupe_tuples:
        :return:
        """

        result = []  # type: List[ColumnReader]
        self._get_col_readers_helper(table=table,
                                     stmt=stmt,
                                     email=email,
                                     dedupe_tuples=dedupe_tuples,
                                     result=result)
        return result

    def _get_col_readers_helper(self,
                                table,  # type: Table
                                stmt,  # type: str
                                email,  # type: str
                                dedupe_tuples,  # type: set
                                result  # type: List[ColumnReader]
                                ):
        # type: (...) -> None
        if not table:
            logging.warn('table does not exist. statement: {}'.format(stmt))
            return

        if isinstance(table, OrTable):
            for table in table.tables:
                self._get_col_readers_helper(table=table,
                                             stmt=stmt,
                                             email=email,
                                             dedupe_tuples=dedupe_tuples,
                                             result=result)
            return

        if self._is_duplicate(table=table, email=email, dedupe_tuples=dedupe_tuples):
            return

        schema = self._get_schema(table)
        if not schema:
            return

        result.append(ColumnReader(database=self._database,
                                   cluster=self._cluster,
                                   schema=schema,
                                   table=table.name,
                                   column='*',
                                   user_email=email))

    def _is_duplicate(self, table, email, dedupe_tuples):
        # type: (Table, str, set) -> bool
        """
        This is to only produce table level usage. TODO: Remove this
        :param table:
        :param email:
        :param dedupe_tuples:
        :return:
        """
        dedupe_tuple = (self._database, self._get_schema(table), table.name, email)
        if dedupe_tuple in dedupe_tuples:
            return True

        dedupe_tuples.add(dedupe_tuple)
        return False

    def get_scope(self):
        # type: () -> str
        return 'transformer.sql_to_tbl_col_usage'

    def _create_schema_by_table_mapping(self):
        # type: () -> dict
        # TODO: Make extractor generic
        table_metadata_extractor = HiveTableMetadataExtractor()
        table_metadata_extractor.init(Scoped.get_scoped_conf(self._conf, table_metadata_extractor.get_scope()))

        table_to_schema = {}
        table_metadata = table_metadata_extractor.extract()
        while table_metadata:
            # TODO: deal with collision
            table_to_schema[table_metadata.name.lower()] = table_metadata.schema_name.lower()
            table_metadata = table_metadata_extractor.extract()
        return table_to_schema

    def _get_schema(self, table):
        # type: (Table) -> Optional[str]
        if table.schema:
            return table.schema.lower()

        return self._tbl_to_schema_mapping.get(table.name.lower())

    def close(self):
        # type: () -> None
        LOGGER.info('Column usage stats: failure: {fail_count} out of total {total_count}'
                    .format(fail_count=SqlToTblColUsageTransformer.failure_counts,
                            total_count=SqlToTblColUsageTransformer.total_counts))



import abc

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.utils.closer import Closer


class Job(Scoped):
    closer = Closer()

    """
    A Databuilder job that represents single work unit.
    """
    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    @abc.abstractmethod
    def launch(self):
        # type: () -> None
        """
        Launch a job
        :return: None
        """
        pass

    def get_scope(self):
        # type: () -> str
        return 'job'



import copy
from collections import namedtuple

from typing import Iterable, Any, Union, Iterator, Dict, Set  # noqa: F401

from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)
from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX

DESCRIPTION_NODE_LABEL = 'Description'


class TagMetadata:
    TAG_NODE_LABEL = 'Tag'
    TAG_KEY_FORMAT = '{tag}'
    TAG_TYPE = 'tag_type'

    def __init__(self,
                 name,  # type: str,
                 tag_type='default',  # type: str
                 ):
        self._name = name
        self._tag_type = tag_type

    @staticmethod
    def get_tag_key(name):
        # type: (str) -> str
        return TagMetadata.TAG_KEY_FORMAT.format(tag=name)


class ColumnMetadata:
    COLUMN_NODE_LABEL = 'Column'
    COLUMN_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}'
    COLUMN_NAME = 'name'
    COLUMN_TYPE = 'type'
    COLUMN_ORDER = 'sort_order'
    COLUMN_DESCRIPTION = 'description'
    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/_description'

    # pair of nodes makes relationship where name of variable represents order of relationship.
    COL_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_COL_RELATION_TYPE = 'DESCRIPTION_OF'

    # Relation between column and tag
    COL_TAG_RELATION_TYPE = 'TAGGED_BY'
    TAG_COL_RELATION_TYPE = 'TAG'

    def __init__(self,
                 name,  # type: str
                 description,  # type: Union[str, None]
                 col_type,  # type: str
                 sort_order,  # type: int
                 tags=None,  # Union[List[str], None]
                 ):
        # type: (...) -> None
        """
        TODO: Add stats
        :param name:
        :param description:
        :param col_type:
        :param sort_order:
        """
        self.name = name
        self.description = description
        self.type = col_type
        self.sort_order = sort_order
        self.tags = tags

    def __repr__(self):
        # type: () -> str
        return 'ColumnMetadata({!r}, {!r}, {!r}, {!r})'.format(self.name,
                                                               self.description,
                                                               self.type,
                                                               self.sort_order)


# Tuples for de-dupe purpose on Database, Cluster, Schema. See TableMetadata docstring for more information
NodeTuple = namedtuple('KeyName', ['key', 'name', 'label'])
RelTuple = namedtuple('RelKeys', ['start_label', 'end_label', 'start_key', 'end_key', 'type', 'reverse_type'])


class TableMetadata(Neo4jCsvSerializable):
    """
    Table metadata that contains columns. It implements Neo4jCsvSerializable so that it can be serialized to produce
    Table, Column and relation of those along with relationship with table and schema. Additionally, it will create
    Database, Cluster, and Schema with relastionships between those.
    These are being created here as it does not make much sense to have different extraction to produce this. As
    database, cluster, schema would be very repititive with low cardinality, it will perform de-dupe so that publisher
    won't need to publish same nodes, relationships.

    This class can be used for both table and view metadata. If it is a View, is_view=True should be passed in.
    """
    TABLE_NODE_LABEL = 'Table'
    TABLE_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}'
    TABLE_NAME = 'name'
    IS_VIEW = 'is_view{}'.format(UNQUOTED_SUFFIX)  # bool value needs to be unquoted when publish to neo4j

    TABLE_DESCRIPTION = 'description'
    TABLE_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/_description'
    TABLE_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_TABLE_RELATION_TYPE = 'DESCRIPTION_OF'

    DATABASE_NODE_LABEL = 'Database'
    DATABASE_KEY_FORMAT = 'database://{db}'
    DATABASE_CLUSTER_RELATION_TYPE = 'CLUSTER'
    CLUSTER_DATABASE_RELATION_TYPE = 'CLUSTER_OF'

    CLUSTER_NODE_LABEL = 'Cluster'
    CLUSTER_KEY_FORMAT = '{db}://{cluster}'
    CLUSTER_SCHEMA_RELATION_TYPE = 'SCHEMA'
    SCHEMA_CLUSTER_RELATION_TYPE = 'SCHEMA_OF'

    SCHEMA_NODE_LABEL = 'Schema'
    SCHEMA_KEY_FORMAT = '{db}://{cluster}.{schema}'
    SCHEMA_TABLE_RELATION_TYPE = 'TABLE'
    TABLE_SCHEMA_RELATION_TYPE = 'TABLE_OF'

    TABLE_COL_RELATION_TYPE = 'COLUMN'
    COL_TABLE_RELATION_TYPE = 'COLUMN_OF'

    TABLE_TAG_RELATION_TYPE = 'TAGGED_BY'
    TAG_TABLE_RELATION_TYPE = 'TAG'

    # Only for deduping database, cluster, and schema (table and column will be always processed)
    serialized_nodes = set()  # type: Set[Any]
    serialized_rels = set()  # type: Set[Any]

    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 name,  # type: str
                 description,  # type: Union[str, None]
                 columns=None,  # type: Iterable[ColumnMetadata]
                 is_view=False,  # type: bool
                 tags=None,  # type: List
                 **kwargs  # type: Dict
                 ):
        # type: (...) -> None
        """
        TODO: Add owners
        :param database:
        :param cluster:
        :param schema_name:
        :param name:
        :param description:
        :param columns:
        :param is_view: Indicate whether the table is a view or not
        :param kwargs: Put additional attributes to the table model if there is any.
        """
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.name = name
        self.description = description
        self.columns = columns if columns else []
        self.is_view = is_view
        self.attrs = None
        self.tags = tags

        if kwargs:
            self.attrs = copy.deepcopy(kwargs)

        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()

    def __repr__(self):
        # type: () -> str
        return 'TableMetadata({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})'.format(self.database,
                                                                                self.cluster,
                                                                                self.schema_name,
                                                                                self.name,
                                                                                self.description,
                                                                                self.columns,
                                                                                self.is_view)

    def _get_table_key(self):
        # type: () -> str
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                     cluster=self.cluster,
                                                     schema=self.schema_name,
                                                     tbl=self.name)

    def _get_table_description_key(self):
        # type: () -> str
        return TableMetadata.TABLE_DESCRIPTION_FORMAT.format(db=self.database,
                                                             cluster=self.cluster,
                                                             schema=self.schema_name,
                                                             tbl=self.name)

    def _get_database_key(self):
        # type: () -> str
        return TableMetadata.DATABASE_KEY_FORMAT.format(db=self.database)

    def _get_cluster_key(self):
        # type: () -> str
        return TableMetadata.CLUSTER_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster)

    def _get_schema_key(self):
        # type: () -> str
        return TableMetadata.SCHEMA_KEY_FORMAT.format(db=self.database,
                                                      cluster=self.cluster,
                                                      schema=self.schema_name)

    def _get_col_key(self, col):
        # type: (ColumnMetadata) -> str
        return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster,
                                                       schema=self.schema_name,
                                                       tbl=self.name,
                                                       col=col.name)

    def _get_col_description_key(self, col):
        # type: (ColumnMetadata) -> str
        return ColumnMetadata.COLUMN_DESCRIPTION_FORMAT.format(db=self.database,
                                                               cluster=self.cluster,
                                                               schema=self.schema_name,
                                                               tbl=self.name,
                                                               col=col.name)

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self):  # noqa: C901
        # type: () -> Iterator[Any]

        table_node = {NODE_LABEL: TableMetadata.TABLE_NODE_LABEL,
                      NODE_KEY: self._get_table_key(),
                      TableMetadata.TABLE_NAME: self.name,
                      TableMetadata.IS_VIEW: self.is_view}
        if self.attrs:
            for k, v in self.attrs.items():
                if k not in table_node:
                    table_node[k] = v
        yield table_node

        if self.description:
            yield {NODE_LABEL: DESCRIPTION_NODE_LABEL,
                   NODE_KEY: self._get_table_description_key(),
                   TableMetadata.TABLE_DESCRIPTION: self.description}

        # Create the table tag node
        if self.tags:
            for tag in self.tags:
                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
                       NODE_KEY: TagMetadata.get_tag_key(tag),
                       TagMetadata.TAG_TYPE: 'default'}

        for col in self.columns:
            yield {
                NODE_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                NODE_KEY: self._get_col_key(col),
                ColumnMetadata.COLUMN_NAME: col.name,
                ColumnMetadata.COLUMN_TYPE: col.type,
                ColumnMetadata.COLUMN_ORDER: col.sort_order}

            if not col.description:
                continue

            yield {
                NODE_LABEL: DESCRIPTION_NODE_LABEL,
                NODE_KEY: self._get_col_description_key(col),
                ColumnMetadata.COLUMN_DESCRIPTION: col.description}

            if not col.tags:
                continue

            for tag in col.tags:
                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
                       NODE_KEY: TagMetadata.get_tag_key(tag),
                       TagMetadata.TAG_TYPE: 'default'}

        # Database, cluster, schema
        others = [NodeTuple(key=self._get_database_key(),
                            name=self.database,
                            label=TableMetadata.DATABASE_NODE_LABEL),
                  NodeTuple(key=self._get_cluster_key(),
                            name=self.cluster,
                            label=TableMetadata.CLUSTER_NODE_LABEL),
                  NodeTuple(key=self._get_schema_key(),
                            name=self.schema_name,
                            label=TableMetadata.SCHEMA_NODE_LABEL)
                  ]

        for node_tuple in others:
            if node_tuple not in TableMetadata.serialized_nodes:
                TableMetadata.serialized_nodes.add(node_tuple)
                yield {
                    NODE_LABEL: node_tuple.label,
                    NODE_KEY: node_tuple.key,
                    'name': node_tuple.name
                }

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self):
        # type: () -> Iterator[Any]

        yield {
            RELATION_START_LABEL: TableMetadata.SCHEMA_NODE_LABEL,
            RELATION_END_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_START_KEY: self._get_schema_key(),
            RELATION_END_KEY: self._get_table_key(),
            RELATION_TYPE: TableMetadata.SCHEMA_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableMetadata.TABLE_SCHEMA_RELATION_TYPE
        }

        if self.description:
            yield {
                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_table_key(),
                RELATION_END_KEY: self._get_table_description_key(),
                RELATION_TYPE: TableMetadata.TABLE_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableMetadata.DESCRIPTION_TABLE_RELATION_TYPE
            }

        if self.tags:
            for tag in self.tags:
                yield {
                    RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                    RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
                    RELATION_START_KEY: self._get_table_key(),
                    RELATION_END_KEY: TagMetadata.get_tag_key(tag),
                    RELATION_TYPE: TableMetadata.TABLE_TAG_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: TableMetadata.TAG_TABLE_RELATION_TYPE,
                }

        for col in self.columns:
            yield {
                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                RELATION_END_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                RELATION_START_KEY: self._get_table_key(),
                RELATION_END_KEY: self._get_col_key(col),
                RELATION_TYPE: TableMetadata.TABLE_COL_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableMetadata.COL_TABLE_RELATION_TYPE
            }

            if not col.description:
                continue

            yield {
                RELATION_START_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_col_key(col),
                RELATION_END_KEY: self._get_col_description_key(col),
                RELATION_TYPE: ColumnMetadata.COL_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: ColumnMetadata.DESCRIPTION_COL_RELATION_TYPE
            }

            if not col.tags:
                continue

            for tag in col.tags:
                yield {
                    RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                    RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
                    RELATION_START_KEY: self._get_table_key(),
                    RELATION_END_KEY: TagMetadata.get_tag_key(tag),
                    RELATION_TYPE: ColumnMetadata.COL_TAG_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: ColumnMetadata.TAG_COL_RELATION_TYPE,
                }

        others = [
            RelTuple(start_label=TableMetadata.DATABASE_NODE_LABEL,
                     end_label=TableMetadata.CLUSTER_NODE_LABEL,
                     start_key=self._get_database_key(),
                     end_key=self._get_cluster_key(),
                     type=TableMetadata.DATABASE_CLUSTER_RELATION_TYPE,
                     reverse_type=TableMetadata.CLUSTER_DATABASE_RELATION_TYPE),
            RelTuple(start_label=TableMetadata.CLUSTER_NODE_LABEL,
                     end_label=TableMetadata.SCHEMA_NODE_LABEL,
                     start_key=self._get_cluster_key(),
                     end_key=self._get_schema_key(),
                     type=TableMetadata.CLUSTER_SCHEMA_RELATION_TYPE,
                     reverse_type=TableMetadata.SCHEMA_CLUSTER_RELATION_TYPE)
        ]

        for rel_tuple in others:
            if rel_tuple not in TableMetadata.serialized_rels:
                TableMetadata.serialized_rels.add(rel_tuple)
                yield {
                    RELATION_START_LABEL: rel_tuple.start_label,
                    RELATION_END_LABEL: rel_tuple.end_label,
                    RELATION_START_KEY: rel_tuple.start_key,
                    RELATION_END_KEY: rel_tuple.end_key,
                    RELATION_TYPE: rel_tuple.type,
                    RELATION_REVERSE_TYPE: rel_tuple.reverse_type
                }



import os
from setuptools import setup, find_packages


__version__ = '1.4.8'


requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')
with open(requirements_path) as requirements_file:
    requirements = requirements_file.readlines()

setup(
    name='amundsen-databuilder',
    version=__version__,
    description='Amundsen Data builder',
    url='https://www.github.com/lyft/amundsendatabuilder',
    maintainer='Lyft',
    maintainer_email='dev@lyft.com',
    packages=find_packages(exclude=['tests*']),
    dependency_links=[],
    install_requires=requirements,
    extras_require={
        ':python_version=="2.7"': ['typing>=3.6'],  # allow typehinting PY2
        'kafka': ['confluent-kafka==1.0.0'],  # To use with Kafka source extractor
    },
)



from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE

from databuilder.models.table_metadata import TableMetadata


class Application(Neo4jCsvSerializable):
    """
    Application-table matching model (Airflow task and table)
    """

    APPLICATION_LABEL = 'Application'
    APPLICATION_KEY_FORMAT = 'application://{cluster}.airflow/{dag}/{task}'
    APPLICATION_URL_NAME = 'application_url'
    APPLICATION_NAME = 'name'
    APPLICATION_ID = 'id'
    APPLICATION_ID_FORMAT = '{dag_id}/{task_id}'
    APPLICATION_DESCRIPTION = 'description'
    APPLICATION_TYPE = 'Airflow'

    APPLICATION_TABLE_RELATION_TYPE = 'GENERATES'
    TABLE_APPLICATION_RELATION_TYPE = 'DERIVED_FROM'

    def __init__(self,
                 task_id,  # type: str
                 dag_id,  # type: str,
                 application_url_template,  # type: str
                 exec_date,  # type: str
                 ):
        # type: (...) -> None
        self.task = task_id

        # todo: need to modify this hack
        self.application_url = application_url_template.format(dag_id=dag_id)
        self.database, self.schema, self.table = task_id.split('.')

        self.dag = dag_id

        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # creates new node
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_table_model_key(self):
        # type: (...) -> str
        # returns formatted string for table name
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                     schema=self.schema,
                                                     tbl=self.table,
                                                     cluster='gold')

    def get_application_model_key(self):
        # type: (...) -> str
        # returns formatting string for application of type dag
        return Application.APPLICATION_KEY_FORMAT.format(cluster='gold',
                                                         dag=self.dag,
                                                         task=self.task)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = []

        results.append({
            NODE_KEY: self.get_application_model_key(),
            NODE_LABEL: Application.APPLICATION_LABEL,
            Application.APPLICATION_URL_NAME: self.application_url,
            Application.APPLICATION_NAME: Application.APPLICATION_TYPE,
            Application.APPLICATION_DESCRIPTION:
                '{app_type} with id {id}'.format(app_type=Application.APPLICATION_TYPE,
                                                 id=Application.APPLICATION_ID_FORMAT.format(dag_id=self.dag,
                                                                                             task_id=self.task)),
            Application.APPLICATION_ID: Application.APPLICATION_ID_FORMAT.format(dag_id=self.dag,
                                                                                 task_id=self.task)
        })

        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relations between application and table nodes
        :return:
        """
        results = [{
            RELATION_START_KEY: self.get_table_model_key(),
            RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_END_KEY: self.get_application_model_key(),
            RELATION_END_LABEL: Application.APPLICATION_LABEL,
            RELATION_TYPE: Application.TABLE_APPLICATION_RELATION_TYPE,
            RELATION_REVERSE_TYPE: Application.APPLICATION_TABLE_RELATION_TYPE
        }]

        return results
